from __future__ import annotations
from typing import Type, TypeVar
from typing_extensions import Self
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalPositionalVectorArray',
    'TemporalPositionalVectorArray',
    'AbstractImplicitTemporalPositionalVectorArray',
    'AbstractParameterizedTemporalPositionalVectorArray',
    'AbstractTemporalPositionalVectorSpace',
    'TemporalPositionalVectorLinearSpace',
    'ExplicitTemporalWcsPositionalVectorArray',
]

TimeT = TypeVar("TimeT", bound=na.ArrayLike)
PositionT = TypeVar("PositionT", bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalPositionalVectorArray(
    na.AbstractPositionalVectorArray,
    na.AbstractTemporalVectorArray,
):

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractTemporalPositionalVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return TemporalPositionalVectorArray

    @property
    def type_matrix(self) -> Type[na.TemporalPositionalMatrixArray]:
        return na.TemporalPositionalMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class TemporalPositionalVectorArray(
    AbstractTemporalPositionalVectorArray,
    na.PositionalVectorArray[PositionT],
    na.TemporalVectorArray
):

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> TemporalPositionalVectorArray:
        return cls(time=scalar, position=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitTemporalPositionalVectorArray(
    AbstractTemporalPositionalVectorArray,
    na.AbstractImplicitPositionalVectorArray,
    na.AbstractImplicitTemporalVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedTemporalPositionalVectorArray(
    AbstractImplicitTemporalPositionalVectorArray,
    na.AbstractParameterizedPositionalVectorArray,
    na.AbstractParameterizedTemporalVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalPositionalVectorSpace(
    AbstractParameterizedTemporalPositionalVectorArray,
    na.AbstractPositionalVectorSpace,
    na.AbstractTemporalVectorSpace
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class TemporalPositionalVectorLinearSpace(
    AbstractTemporalPositionalVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ExplicitTemporalWcsPositionalVectorArray(
    AbstractImplicitTemporalPositionalVectorArray,
    na.AbstractWcsVector,
):
    time: na.AbstractScalar = dataclasses.MISSING
    crval: na.AbstractPositionalVectorArray = dataclasses.MISSING
    crpix: na.AbstractCartesianNdVectorArray = dataclasses.MISSING
    cdelt: na.AbstractPositionalVectorArray = dataclasses.MISSING
    pc: na.AbstractPositionalMatrixArray = dataclasses.MISSING
    shape_wcs: dict[str, int] = dataclasses.MISSING

    @property
    def _components_explicit(self) -> dict[str, na.ArrayLike]:
        return dict(time=self.time)



from __future__ import annotations
from typing import Type

import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalPositionalMatrixArray',
    'TemporalPositionalMatrixArray',
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalPositionalMatrixArray(
    na.AbstractTemporalPositionalVectorArray,
):

    @property
    def type_abstract(self) -> Type[AbstractTemporalPositionalMatrixArray]:
        return AbstractTemporalPositionalMatrixArray

    @property
    def type_explicit(self) -> Type[TemporalPositionalMatrixArray]:
        return TemporalPositionalMatrixArray

    @property
    def type_vector(self) -> Type[na.TemporalPositionalVectorArray]:
        return na.TemporalPositionalVectorArray

    @property
    def determinant(self) -> na.ScalarLike:     # pragma: nocover
        return NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class TemporalPositionalMatrixArray(
    na.TemporalPositionalVectorArray,
    AbstractTemporalPositionalMatrixArray,
    na.AbstractExplicitMatrixArray,
):
    pass



from __future__ import annotations
import functools
from typing import TypeVar, Generic, Type, ClassVar, Sequence, Callable, Collection, Any, Literal
from typing_extensions import Self
import abc
import dataclasses
import numpy as np
import astropy.units as u
import astropy.visualization
import named_arrays as na
import itertools
import collections

__all__ = [
    "InputValueError",
    "AbstractFunctionArray",
    "FunctionArray",
    "AbstractFunctionArray",
    "AbstractPolynomialFunctionArray",
    "PolynomialFitFunctionArray",
]

InputsT = TypeVar("InputsT", bound=na.AbstractArray)
OutputsT = TypeVar("OutputsT", bound=na.AbstractArray)


class InputValueError(ValueError):
    """
    Exception raised when the inputs of two functions do not match
    """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractFunctionArray(
    na.AbstractArray,
):

    @property
    @abc.abstractmethod
    def inputs(self) -> na.AbstractArray:
        """
        The arrays representing the inputs to this function.
        """

    @property
    @abc.abstractmethod
    def outputs(self) -> na.AbstractArray:
        """
        The arrays representing the outputs of this function.
        """

    __named_array_priority__: ClassVar[float] = (100 * na.AbstractVectorArray.__named_array_priority__)

    @property
    def axes_center(self) -> tuple(str):
        """
        Return keys corresponding to all input axes representing bin centers
        """
        axes_center = tuple()
        input_shape = self.inputs.shape
        output_shape = self.outputs.shape

        for axis in self.axes:
            if axis in input_shape:
                if axis in output_shape:
                    if input_shape[axis] == output_shape[axis]:
                        axes_center += (axis,)
                    else:
                        if input_shape[axis] == 1 or output_shape[axis] == 1:
                            axes_center += (axis,)
                        elif input_shape[axis] != output_shape[axis] + 1: # pragma: no cover
                            raise ValueError(
                                f"Output {axis=} dimension, {output_shape[axis]=}, must either match input axis dimension  {input_shape[axis]=}, (representing"
                                " bin centers) or exceed by one (representing bin vertices)."
                            )
                else:
                    axes_center += (axis,)
            else:
                axes_center += (axis,)

        return axes_center

    @property
    def axes_vertex(self) -> tuple(str):
        """
        Return keys corresponding to all input axes representing bin vertices
        """
        axes_vertex = tuple()

        for axis in self.axes:
            if axis not in self.axes_center:
                axes_vertex += (axis,)

        return axes_vertex

    @property
    def type_abstract(self) -> Type[AbstractFunctionArray]:
        return AbstractFunctionArray

    @property
    def type_explicit(self) -> Type[FunctionArray]:
        return type(self)

    @property
    def value(self) -> FunctionArray:
        exp = self.explicit
        return exp.replace(
            outputs=self.outputs.value,
        )

    @property
    def broadcasted(self) -> FunctionArray:

        exp = self.explicit

        broadcasted_shape_outputs = exp.shape
        broadcasted_shape_inputs = broadcasted_shape_outputs.copy()
        
        axes_vertex = exp.axes_vertex
        inputs = exp.inputs
        for key in broadcasted_shape_outputs:
            axes_vertex = axes_vertex
            if key in axes_vertex:  
                broadcasted_shape_inputs[key] = inputs.shape[key]

        return exp.replace(
            inputs=exp.inputs.broadcast_to(broadcasted_shape_inputs),
            outputs=exp.outputs.broadcast_to(broadcasted_shape_outputs),
        )

    def astype(
            self,
            dtype: str | np.dtype | type,
            order: str = 'K',
            casting='unsafe',
            subok: bool = True,
            copy: bool = True,
    ) -> FunctionArray:

        exp = self.explicit

        return exp.replace(
            inputs=exp.inputs,
            outputs=exp.outputs.astype(
                dtype=dtype,
                order=order,
                casting=casting,
                subok=subok,
                copy=copy,
            )
        )

    def to(
        self,
        unit: u.UnitBase,
        equivalencies: None | list[tuple[u.Unit, u.Unit]] = [],
        copy: bool = True,
    ) -> FunctionArray:
        exp = self.explicit
        return exp.replace(
            outputs=exp.outputs.to(
                unit=unit,
                equivalencies=equivalencies,
                copy=copy,
            ),
        )

    @property
    def length(self) -> FunctionArray:
        exp = self.explicit
        return exp.replace(
            outputs=exp.outputs.length,
        )

    def add_axes(self, axes: str | Sequence[str]) -> FunctionArray:
        exp = self.explicit
        return exp.replace(
            outputs=exp.outputs.add_axes(axes),
        )

    def combine_axes(
            self,
            axes: Sequence[str] = None,
            axis_new: str = None,
    ) -> FunctionArray:

        self = self.explicit

        axes = self.axes if axes is None else axes

        if not set(axes).issubset(set(self.axes)):
            raise ValueError(f"Axes {set(axes) - set(self.axes)} not in {self.axes}")

        for axis in axes:
            if axis in self.axes_vertex:
                raise ValueError(f"Axis {axis} describes input vertices and cannot be used in combine_axes.")

        inputs = self.inputs
        outputs = self.outputs
        inputs_shape_broadcasted = inputs.shape
        outputs_shape_broadcasted = outputs.shape

        shape = self.shape
        for axis in axes:
            if axis not in inputs_shape_broadcasted:
                inputs_shape_broadcasted[axis] = shape[axis]
            if axis not in outputs_shape_broadcasted:
                outputs_shape_broadcasted[axis] = shape[axis]

        inputs = na.broadcast_to(inputs, inputs_shape_broadcasted)
        outputs = na.broadcast_to(outputs, outputs_shape_broadcasted)

        return self.replace(
            inputs=inputs.combine_axes(axes=axes, axis_new=axis_new),
            outputs=outputs.combine_axes(axes=axes, axis_new=axis_new),
        )

    def __call__(
        self,
        inputs: na.AbstractArray,
        axis: None | str | tuple[str] = None,
        method: Literal['multilinear', 'conservative'] = 'multilinear',
        weights: None | tuple[na.AbstractScalar, dict[str, int], dict[str, int]] = None,
    ) -> Self:
        return self.regrid(
            inputs=inputs,
            axis=axis,
            method=method,
            weights=weights,
        )

    def regrid(
        self,
        inputs: na.AbstractArray,
        axis: None | str | tuple[str] = None,
        method: Literal['multilinear', 'conservative'] = 'multilinear',
        weights: None | tuple[na.AbstractScalar, dict[str, int], dict[str, int]] = None,
    ) -> Self:
        """
        Resample this function array onto a new set of input coordinates
        using :func:`named_arrays.regridding.regrid`.

        Parameters
        ----------
        inputs
            The new input coordinates on which to resample the outputs.
        axis
            The logical axes of the input over which to resample.
        method
            The resampling method to use.
        weights
            Optional weights which can be computed in advance using :meth:`weights`
            to greatly speed repeated resampling of the same `inputs`.

        See Also
        --------
        :meth:`weights`: If you need to resample the same coordinates more than once.
        """

        _self = self.explicit

        inputs_new, _, _ = _self._normalize__regrid__args(
            inputs=inputs,
            axis=axis,
        )

        if weights is not None:
            _weights, shape_input, shape_output = weights

        else:
            _weights, shape_input, shape_output = _self.weights(
                inputs=inputs,
                axis=axis,
                method=method,
            )

        outputs_new = na.regridding.regrid_from_weights(
            weights=_weights,
            shape_input=shape_input,
            shape_output=shape_output,
            values_input=_self.outputs,
        )

        final_coordinates_dict = {}

        if isinstance(inputs, na.AbstractVectorArray) and isinstance(_self.inputs, na.AbstractVectorArray):

            for c in _self.inputs.cartesian_nd.components:
                if inputs.cartesian_nd.components[c] is None:  # pragma: no cover
                    final_coordinates_dict[c] = _self.inputs.cartesian_nd.components[c]
                else:
                    final_coordinates_dict[c] = inputs.cartesian_nd.components[c]

            return _self.replace(
                inputs=_self.inputs.type_explicit.from_cartesian_nd(
                    array=na.CartesianNdVectorArray(final_coordinates_dict),
                    like=_self.inputs
                ),
                outputs=outputs_new,
            )
        else:
            return _self.replace(
                inputs=inputs_new,
                outputs=outputs_new,
            )

    def weights(
        self,
        inputs: na.AbstractArray,
        axis: None | str | tuple[str] = None,
        method: Literal['multilinear', 'conservative'] = 'multilinear',
    ) -> tuple[na.AbstractScalar, dict[str, int], dict[str, int]]:
        """
        Compute the resampling weights of this array using
        :func:`named_arrays.regridding.weights`.
        The output of this method is designed to be used by :meth:`regrid`.

        Parameters
        ----------
        inputs
            The new input coordinates on which to resample the outputs.
        axis
            The logical axes of the input over which to resample.
        method
            The resampling method to use.

        See Also
        --------
        :meth:`regrid`: A method designed to use these weights.
        """

        coordinates_new, coordinates_old, axis_input = self._normalize__regrid__args(
            inputs=inputs,
            axis=axis,
        )

        return na.regridding.weights(
            coordinates_input=coordinates_old,
            coordinates_output=coordinates_new,
            axis_input=axis_input,
            method=method
        )

    def _normalize__regrid__args(
        self,
        inputs: na.AbstractArray,
        axis: None | str | tuple[str],
    ):

        inputs_old = self.inputs
        inputs_new = inputs

        if isinstance(inputs_new, na.AbstractVectorArray):
            new_input_components = inputs_new.cartesian_nd.components
        else:
            new_input_components = dict(_dummy=inputs_new.explicit)

        if isinstance(inputs_old, na.AbstractVectorArray):
            old_input_components = inputs_old.cartesian_nd.components
        else:
            old_input_components = dict(_dummy=inputs_old.explicit)

        # broadcast new inputs against value to be interpolated
        if axis is None:
            axis = na.shape_broadcasted(
                new_input_components[c] for c in new_input_components if new_input_components[c] is not None
            )
            axis = tuple(axis)

        # check physical(vector) dimensions of each input match
        if new_input_components.keys() == old_input_components.keys():

            coordinates_new = {}
            coordinates_old = {}
            for c in new_input_components:
                component = new_input_components[c]
                if component is not None:
                    # if input components logical axes do not include interp axes, skip
                    if not set(axis).isdisjoint(component.axes):
                        coordinates_new[c] = component
                        coordinates_old[c] = old_input_components[c]

                else:
                    # check if uninterpolated physical axes vary along interpolation axes
                    if not set(axis).isdisjoint(old_input_components[c].axes):  # pragma: no cover
                        raise ValueError(
                            f"If a component is marked separable using `None`, its shape, {old_input_components[c].axes},"
                            f"should be disjoint from the interpolated axes, {axis}.",
                        )

        else:
            raise ValueError('Physical axes of new and old inputs must match.')  # pragma: no cover

        if isinstance(inputs_new, na.AbstractVectorArray):
            coordinates_new = na.CartesianNdVectorArray.from_components(coordinates_new)
        else:
            coordinates_new = coordinates_new['_dummy']

        if isinstance(inputs_old, na.AbstractVectorArray):
            coordinates_old = na.CartesianNdVectorArray.from_components(coordinates_old)
        else:
            coordinates_old = coordinates_old['_dummy']

        return coordinates_new, coordinates_old, axis

    def cell_centers(
        self,
        axis: None | str | Sequence[str] = None,
        random: bool = False,
    ) -> na.AbstractExplicitArray:
        exp = self.explicit
        return exp.replace(
            inputs=exp.inputs.cell_centers(axis, random=random),
            outputs=exp.outputs.cell_centers(axis, random=random),
        )

    def to_string_array(
        self,
        format_value: str = "%.2f",
        format_unit: str = "latex_inline",
        pad_unit: str = r"$\,$",
    ):
        exp = self.explicit
        kwargs = dict(
            format_value=format_value,
            format_unit=format_unit,
            pad_unit=pad_unit,
        )
        return exp.replace(
            inputs=na.as_named_array(exp.inputs).to_string_array(**kwargs),
            outputs=na.as_named_array(exp.outputs).to_string_array(**kwargs),
        )

    def _getitem(
            self,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray | na.AbstractFunctionArray,
    ) -> FunctionArray:

        array = self.explicit
        inputs = array.inputs
        outputs = array.outputs

        if isinstance(item, na.AbstractArray):
            if isinstance(item, na.AbstractFunctionArray):
                if not np.all(item.inputs == array.inputs):
                    raise ValueError("item inputs do not match function inputs.")

                axes = set(inputs.axes) - set(array.axes_center)
                inputs = inputs.cell_centers(axis=axes)
                item_inputs = item.outputs
                item_outputs = item.outputs
            else:
                item_inputs = item_outputs = item

            shape_item_inputs = item_inputs.shape
            shape_item_outputs = item_outputs.shape

            inputs = na.broadcast_to(inputs, na.broadcast_shapes(inputs.shape, shape_item_inputs))
            outputs = na.broadcast_to(outputs, na.broadcast_shapes(outputs.shape, shape_item_outputs))

        elif isinstance(item, dict):

            if not set(item).issubset(array.axes): # pragma: no cover
                raise ValueError(f"item contains axes {set(item) - set(array.axes)} that does not exist in {set(array.axes)}")

            item_inputs = dict()
            item_outputs = dict()
            for ax in item:
                item_ax = item[ax]
                if isinstance(item_ax, na.AbstractFunctionArray):
                    item_inputs[ax] = item_ax.inputs
                    item_outputs[ax] = item_ax.outputs
                else:
                    axes_center = array.axes_center
                    if ax in axes_center:
                        #can't assume center ax is in both outputs and inputs
                        if ax in inputs.shape:
                            item_inputs[ax] = item_ax
                        if ax in outputs.shape:
                            item_outputs[ax] = item_ax
                    axes_vertex = array.axes_vertex
                    if ax in axes_vertex:
                        if isinstance(item_ax, int):
                            item_outputs[ax] = slice(item_ax, item_ax + 1)
                            item_inputs[ax] = slice(item_ax, item_ax + 2)
                        elif isinstance(item_ax, slice):
                            item_outputs[ax] = item_ax
                            if item_ax.start is None and item_ax.stop is None:
                                item_inputs[ax] = item_ax
                            else:
                                if item_ax.stop is not None:
                                    item_inputs[ax] = slice(item_ax.start, item_ax.stop + 1)
                                else:
                                    item_inputs[ax] = slice(item_ax.start, None)
                        else:
                            return NotImplemented

        else:
            return NotImplemented

        return array.replace(
            inputs=inputs[item_inputs],
            outputs=outputs[item_outputs],
        )

    def _getitem_reversed(
            self,
            array: na.AbstractArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractFunctionArray
    ):
        if isinstance(array, (na.AbstractScalar, na.AbstractVectorArray)):
            array = na.FunctionArray(array, array)
        else:
            return NotImplemented
        return array._getitem(item)

    def __bool__(self) -> bool:
        result = super().__bool__()
        return result and bool(self.outputs)

    def __mul__(self, other: na.ArrayLike | u.UnitBase) -> FunctionArray:
        if isinstance(other, u.UnitBase):
            exp = self.explicit
            return exp.replace(
                outputs=exp.outputs * other,
            )
        else:
            return super().__mul__(other)

    def __lshift__(self, other: na.ArrayLike | u.UnitBase) -> FunctionArray:
        if isinstance(other, u.UnitBase):
            exp = self.explicit
            return exp.replace(
                outputs=self.outputs << other,
            )
        else:
            return super().__lshift__(other)

    def __truediv__(self, other: na.ArrayLike | u.UnitBase) -> FunctionArray:
        if isinstance(other, u.UnitBase):
            exp = self.explicit
            return exp.replace(
                outputs=self.outputs / other,
            )
        else:
            return super().__truediv__(other)

    def __array_matmul__(
            self,
            x1: float | u.Quantity | FunctionArray,
            x2: float | u.Quantity | FunctionArray,
            out: None | na.FunctionArray = None,
            **kwargs,
    ) -> FunctionArray:

        if isinstance(x1, na.AbstractArray):
            if isinstance(x1, AbstractFunctionArray):
                inputs_1 = x1.inputs
                outputs_1 = x1.outputs
            else:
                return NotImplemented
        else:
            inputs_1 = None
            outputs_1 = x1

        if isinstance(x2, na.AbstractArray):
            if isinstance(x2, AbstractFunctionArray):
                inputs_2 = x2.inputs
                outputs_2 = x2.outputs
            else:
                return NotImplemented
        else:
            inputs_2 = None
            outputs_2 = x2

        if inputs_1 is not None:
            inputs = inputs_1
            if inputs_2 is not None:
                if np.any(inputs_1 != inputs_2):
                    raise InputValueError("`x1.inputs` and `x2.inputs` must be equal")
            else:
                pass
        else:
            if inputs_2 is not None:
                inputs = inputs_2
            else:
                return NotImplemented

        if out is not None:
            if not np.all(inputs == out.inputs):
                raise InputValueError("`out.inputs` must be equal to `x1.inputs` and `x2.inputs`")
        result = self.explicit.replace(
            inputs=inputs,
            outputs=np.matmul(
                outputs_1,
                outputs_2,
                out=out.outputs if out is not None else out,
                **kwargs,
            ),
        )

        if out is not None:
            result = out

        return result

    def __array_ufunc__(
            self,
            function: np.ufunc,
            method: str,
            *inputs,
            **kwargs,
    ) -> None | na.AbstractArray | tuple[na.AbstractArray, ...]:

        result = super().__array_ufunc__(function, method, *inputs, **kwargs)
        if result is not NotImplemented:
            return result

        nout = function.nout

        inputs_inputs = []
        inputs_outputs = []
        for inp in inputs:
            if isinstance(inp, na.AbstractArray):
                if isinstance(inp, AbstractFunctionArray):
                    inputs_inputs.append(inp.inputs)
                    inputs_outputs.append(inp.outputs)
                else:
                    if not inp.shape:
                        inputs_outputs.append(inp)
                    else:
                        return NotImplemented
            else:
                inputs_outputs.append(inp)

        for inputs_input in inputs_inputs[1:]:
            if np.any(inputs_input != inputs_inputs[0]):
                raise InputValueError(f"all inputs to {function} must have the same coordinates")

        if "out" in kwargs:
            out = kwargs.pop("out")
            outputs_out = list()
            for o in out:
                if o is not None:
                    if isinstance(o, self.type_explicit):
                        outputs_o = o.outputs
                    else:
                        raise ValueError(
                            f"each element of `out` must be an instance of {self.type_explicit}, got {type(o)}"
                        )
                else:
                    outputs_o = None
                outputs_out.append(outputs_o)
            if nout == 1:
                outputs_out = outputs_out[0]
            else:
                outputs_out = tuple(outputs_out)

            kwargs["out"] = outputs_out
        else:
            out = (None,) * nout

        if "where" in kwargs:
            where = kwargs.pop("where")
            if isinstance(where, na.AbstractArray):
                if isinstance(where, AbstractFunctionArray):
                    if np.any(where.inputs != inputs_inputs[0]):
                        raise InputValueError(f"`where.inputs` must match the rest of the inputs")
                    kwargs["where"] = where.outputs
                else:
                    return NotImplemented
            else:
                kwargs["where"] = where

        inputs_result = inputs_inputs[0]
        outputs_result = getattr(function, method)(*inputs_outputs, **kwargs)

        if nout == 1:
            outputs_result = (outputs_result,)

        result = list(
            self.explicit.replace(
                inputs=inputs_result,
                outputs=outputs_result[i],
            )
            for i in range(nout)
        )

        for i in range(nout):
            if out[i] is not None:
                out[i].inputs = result[i].inputs
                out[i].outputs = result[i].outputs
                result[i] = out[i]

        if nout == 1:
            result = result[0]
        else:
            result = tuple(result)

        return result

    def __array_function__(
            self,
            func: Callable,
            types: Collection,
            args: tuple,
            kwargs: dict[str, Any],
    ):
        result = super().__array_function__(func=func, types=types, args=args, kwargs=kwargs)
        if result is not NotImplemented:
            return result

        from . import function_array_functions

        if func in function_array_functions.DEFAULT_FUNCTIONS:
            return function_array_functions.array_function_default(func, *args, **kwargs)

        if func in function_array_functions.PERCENTILE_LIKE_FUNCTIONS:
            return function_array_functions.array_function_percentile_like(func, *args, **kwargs)

        if func in function_array_functions.ARG_REDUCE_FUNCTIONS:
            return function_array_functions.array_function_arg_reduce(func, *args, **kwargs)

        if func in function_array_functions.STACK_LIKE_FUNCTIONS:
            return function_array_functions.array_function_stack_like(func, *args, **kwargs)

        if func in function_array_functions.HANDLED_FUNCTIONS:
            return function_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def __named_array_function__(self, func, *args, **kwargs):
        result = super().__named_array_function__(func, *args, **kwargs)
        if result is not NotImplemented:
            return result

        from . import function_named_array_functions

        if func in function_named_array_functions.ASARRAY_LIKE_FUNCTIONS:
            return function_named_array_functions.asarray_like(func=func, *args, **kwargs)

        if func in function_named_array_functions.NDFILTER_FUNCTIONS:
            return function_named_array_functions.ndfilter(func, *args, **kwargs)

        if func in function_named_array_functions.HANDLED_FUNCTIONS:
            return function_named_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def interp_linear(
            self,
            item: dict[str, na.AbstractArray],
    ) -> FunctionArray:
        raise NotImplementedError


    def pcolormesh(
            self,
            axs: np.ndarray,
            input_component_x: str,
            input_component_y: str,
            input_component_row: None | str = None,
            input_component_column: None | str = None,
            index: None | dict[str, int] = None,
            output_component_color: None | str = None,
            **kwargs: Any,
    ):
        """
        Plot a :class:`FunctionArray` via :func:`matplotlib.pyplot.pcolormesh`.

        :func:`FunctionArray.pcolormesh` takes in an axes object, or array of axes objects, along with components to
        be plotted along the x and y plot axes (:attr:`input_component_x` and :attr:`input_component_y`). Additional
        components can be tiled along subplot row/column and are specified in :attr:`input_component_row` and
        :attr:`input_component_column`.

        .. jupyter-execute::

            import named_arrays as na
            import numpy as np
            import astropy.units as u
            import matplotlib.pyplot as plt

            position = na.Cartesian2dVectorLinearSpace(
                start=-10,
                stop=10,
                axis=na.Cartesian2dVectorArray(
                    x='position_x',
                    y='position_y',
                ),
                num=21,
            ) * u.m

            x_width = 5 * u.m
            y_width = 2 * u.m
            velocity = 1 * u.m/u.s
            time = na.ScalarLinearSpace(
                start=0 * u.s,
                stop=3 * u.s,
                num=4,
                axis='time'
            )

            intensity = np.exp(-(((position.x + velocity*time)/x_width) ** 2 + ((position.y + 2*velocity*time)/y_width)** 2))
            scene = na.FunctionArray(
                inputs=position,
                outputs=intensity,
            )

            fig, axs = plt.subplots(
                nrows=scene.outputs.shape['time'],
                squeeze=False,
                sharex=True,
                subplot_kw=dict(aspect='equal'),
            )
            scene.pcolormesh(
                axs=axs,
                input_component_x='x',
                input_component_y='y',
                input_component_row='time',
            )
        """

        if axs.ndim == 1:
            if input_component_row is not None:
                axs = axs[..., np.newaxis]
            if input_component_column is not None:
                axs = axs[np.newaxis, ...]

        axs = na.ScalarArray(
            ndarray=axs,
            axes=('row', 'column')
        )

        if index is None:
            index = dict()

        with astropy.visualization.quantity_support():
            for index_subplot in axs.ndindex():

                index_final = index.copy()
                if input_component_row is not None:
                    index_final[input_component_row] = index_subplot['row']
                if input_component_column is not None:
                    index_final[input_component_column] = index_subplot['column']

                inp = self[index_final].inputs.cartesian_nd

                inp_x = inp.components[input_component_x].ndarray
                inp_y = inp.components[input_component_y].ndarray

                out = self[index_final].outputs
                if output_component_color is not None:
                    out = out.components[output_component_color]

                ax = axs[index_subplot].ndarray
                ax.pcolormesh(
                    inp_x,
                    inp_y,
                    out.ndarray,
                    shading='auto',
                    **kwargs,
                )

                if index_subplot['row'] == axs.shape['row'] - 1:
                    if isinstance(inp_x, u.Quantity):
                        ax.set_xlabel(f'{input_component_x} ({inp_x.unit})')
                    else:
                        ax.set_xlabel(f'{input_component_x}')
                else:
                    ax.set_xlabel(None)

                if index_subplot['column'] == 0:
                    if isinstance(inp_y, u.Quantity):
                        ax.set_ylabel(f'{input_component_y} ({inp_y.unit})')
                    else:
                        ax.set_ylabel(f'{input_component_y}')
                else:
                    ax.set_ylabel(None)

                if input_component_column is not None:
                    if input_component_column in inp.components:
                        if index_subplot['row'] == 0:
                            inp_column = inp.components[input_component_column]
                            ax.text(
                                x=0.5,
                                y=1.01,
                                s=f'{inp_column.mean().ndarray.value:0.03f} {inp_column.unit:latex_inline}',
                                transform=ax.transAxes,
                                ha='center',
                                va='bottom'
                            )

                if input_component_row is not None:
                    if input_component_row in inp.components:
                        if index_subplot['column'] == axs.shape['column'] - 1:
                            inp_row = inp.components[input_component_row]
                            ax.text(
                                x=1.01,
                                y=0.5,
                                s=f'{inp_row.mean().ndarray.value:0.03f} {inp_row.unit:latex_inline}',
                                transform=ax.transAxes,
                                va='center',
                                ha='left',
                                rotation=-90,
                            )


@dataclasses.dataclass(eq=False, repr=False)
class FunctionArray(
    AbstractFunctionArray,
    na.AbstractExplicitArray,
    Generic[InputsT, OutputsT],
):
    inputs: InputsT = 0
    outputs: OutputsT = 0

    @classmethod
    def from_scalar_array(
            cls: type[Self],
            a: float | u.Quantity | na.AbstractScalarArray,
            like: None | Self = None,
    ) -> Self:

        self = super().from_scalar_array(a=a, like=like)

        if like is None:
            self.inputs = a
            self.outputs = a
        else:
            if isinstance(like.inputs, na.AbstractArray):
                self.inputs = like.inputs.from_scalar_array(a=a, like=like.inputs)
            else:
                self.inputs = a

            if isinstance(like.outputs, na.AbstractArray):
                self.outputs = like.outputs.from_scalar_array(a=a, like=like.outputs)
            else:
                self.outputs = a

        return self

    @property
    def axes(self) -> tuple[str, ...]:
        return tuple(self.inputs.shape | self.outputs.shape)

    @property
    def shape(self) -> dict[str, int]:

        axes_center = self.axes_center
        axes_vertex = self.axes_vertex

        outputs_shape = self.outputs.shape
        inputs_shape = self.inputs.shape

        outputs_shape_center = {ax: outputs_shape[ax] for ax in outputs_shape if ax in axes_center}
        inputs_shape_center = {ax: inputs_shape[ax] for ax in inputs_shape if ax in axes_center}
        outputs_shape_vertex = {ax: outputs_shape[ax] for ax in outputs_shape if ax in axes_vertex}

        shape = na.broadcast_shapes(
            outputs_shape_center,
            outputs_shape_vertex,
            inputs_shape_center,
        )

        return shape

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def size(self) -> int:
        return int(np.array(tuple(self.shape.values())).prod())

    @property
    def explicit(self) -> FunctionArray:
        return self.replace(
            inputs=na.explicit(self.inputs),
            outputs=na.explicit(self.outputs),
        )

    def __setitem__(
            self,
            item: dict[str, int | slice | na.AbstractScalar | na.AbstractFunctionArray] | na.AbstractFunctionArray,
            value: float | u.Quantity | na.FunctionArray,
    ):

        if isinstance(item, na.AbstractFunctionArray):
            if not np.all(item.inputs == self.inputs):
                raise ValueError("boolean advanced index does not have the same inputs as the array")

            item_inputs = item.outputs
            item_outputs = item.outputs

        elif isinstance(item, dict):

            item_inputs = dict()
            item_outputs = dict()
            for ax in item:
                item_ax = item[ax]
                if isinstance(item_ax, na.AbstractFunctionArray):
                    item_inputs[ax] = item_ax.inputs
                    item_outputs[ax] = item_ax.outputs
                else:
                    item_inputs[ax] = item_outputs[ax] = item_ax
        else:
            raise TypeError(
                f"`item` must be an instance of `{dict.__name__}`, or `{na.AbstractFunctionArray.__name__}`, "
                f"got `{type(item)}`"
            )

        if isinstance(value, na.AbstractArray):
            if isinstance(value, na.AbstractFunctionArray):
                value_inputs = value.inputs
                value_outputs = value.outputs

                # maybe this should only set to None vertex axes only
                axes_vertex = self.axes_vertex
                for ax in value_inputs.shape:
                    axes_vertex = axes_vertex
                    if ax in axes_vertex:
                        value_inputs = None

            else:
                if value.shape:
                    raise ValueError(
                        f"if `value` is an instance of `{na.AbstractArray.__name__}`, "
                        f"but not an instance of `{na.AbstractFunctionArray.__name__}`, "
                        f"`value.shape` should be empty, got {value.shape}"
                    )
                else:
                    value_inputs = None
                    value_outputs = value
        else:
            value_inputs = None
            value_outputs = value

        if value_inputs is not None:
            self.inputs[item_inputs] = value_inputs

        self.outputs[item_outputs] = value_outputs


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPolynomialFunctionArray(
    AbstractFunctionArray,
):
    @property
    @abc.abstractmethod
    def coefficients(self) -> na.AbstractVectorArray | na.AbstractMatrixArray:
        """
        A vector or matrix representing the coefficients of the polynomial.

        If this function is scalar-valued, :attr:`coefficients` should be a vector,
        and if this function is vector-valued, :attr`coefficients` should be a matrix.
        """

    @property
    @abc.abstractmethod
    def degree(self) -> int:
        """degree of the polynomial"""

    @property
    @abc.abstractmethod
    def components_polynomial(self) -> None | str | Sequence[str]:
        """The components of the input that this polynomial depends on."""

    @property
    @abc.abstractmethod
    def axis_polynomial(self) -> None | str | Sequence[str]:
        """the logical axes along which this polynomial is distributed"""

    @abc.abstractmethod
    def design_matrix(
        self,
        inputs: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    ) -> na.AbstractVectorArray:
        """
        The `design matrix <https://en.wikipedia.org/wiki/Design_matrix>`_
        corresponding to the given inputs.

        Note that while this is `called` a matrix, this function returns a vector
        since the rows are independent observations, and this is concept is
        already captured by the logical axes in the array.

        Parameters
        ----------
        inputs
            the set of independent variables to convert into the design matrix
        """

    def __call__(
        self,
        inputs: float | u.Quantity | na.ScalarArray | na.AbstractVectorArray,
    ) -> na.AbstractFunctionArray:
        return na.FunctionArray(
            inputs, self.design_matrix(inputs) @ self.coefficients
        )

    @property
    def predictions(self) -> OutputsT:
        """
        The outputs of the polynomial model given :attr:`inputs`.
        Equivalent to ``self(self.inputs).outputs``.
        """
        return self(self.inputs).outputs


@dataclasses.dataclass(eq=False, repr=False)
class PolynomialFitFunctionArray(
    FunctionArray,
    AbstractPolynomialFunctionArray,
):
    """
    A :class:`named_arrays.PolynomialFitFunctionArray` carries the independent variables, inputs, and dependent variables, outputs,
    of a discrete function, and a linear least squares polynomial fit of specified degree to that function.

    Parameters
    ----------
    inputs
        the set of independent variables
    outputs
        the set of dependent variables
    degree
        the degree of the polynomial
    components_polynomial
        the components used in the polynomial fit
    axis_polynomial
        the logical axis of the polynomial fit

    Examples
    --------

    .. nblinkgallery::
        :caption: Relevant Tutorials
        :name: rst-link-gallery

        ../tutorials/PolynomialFunctionArray
    """

    degree: int = None
    components_polynomial: None | str | Sequence[str] = None
    axis_polynomial: None | str | Sequence[str] = None

    @functools.cached_property
    def coefficients(self) -> na.AbstractVectorArray | na.AbstractMatrixArray:
        d = self.design_matrix(self.inputs)
        dTd = self._outer(d, d, self.axis_polynomial)
        dTo = self._outer(d, self.outputs, self.axis_polynomial)

        return dTd.inverse @ dTo

    def design_matrix(
        self,
        inputs: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    ) -> na.AbstractVectorArray:

        design_matrix = {}

        if isinstance(inputs, na.AbstractScalar):
            inputs = na.CartesianNdVectorArray({"dummy": inputs})
        inputs = inputs.cartesian_nd.broadcasted.components

        components = self.components_polynomial

        if components is None:
            components = tuple(inputs)
        elif isinstance(components, str):
            components = (components,)

        inputs = {c: inputs[c] for c in components}

        for i in range(self.degree + 1):
            combinations = itertools.combinations_with_replacement(
                inputs, i
            )
            for combination in combinations:
                key = "*".join(combination)
                design_matrix[key] = 1
                for k in combination:
                    design_matrix[key] = design_matrix[key] * inputs[k]

        design_matrix = na.CartesianNdVectorArray(design_matrix)

        return design_matrix

    @classmethod
    def _outer(cls, v1, v2, axis):
        v1_T_v2_components = {}

        if isinstance(v1, na.AbstractVectorArray):
            v1_broadcasted = v1.broadcasted.components
            if isinstance(v2, na.AbstractVectorArray):
                for c1 in v1_broadcasted:
                    v2_broadcasted = v2.broadcasted.components
                    row_components = {}
                    for c2 in v2_broadcasted:
                        row_components[c2] = (
                                v1_broadcasted[c1] * v2_broadcasted[c2]
                        ).sum(axis=axis)
                    v1_T_v2_components[c1] = v2.type_explicit.from_components(row_components)
                v1_T_v2 = v1.type_matrix.from_components(v1_T_v2_components)

            else:
                for c1 in v1_broadcasted:
                        row_components = (v1_broadcasted[c1] * v2).sum(axis=axis)
                        v1_T_v2_components[c1] = row_components
                v1_T_v2 = v1.type_explicit.from_components(v1_T_v2_components)

        return v1_T_v2






from __future__ import annotations
from typing import TypeVar, Generic, ClassVar, Type, Sequence, Callable, Collection, Any, Union, cast, Dict
from typing_extensions import Self
import abc
import dataclasses
import numpy as np
import numpy.typing as npt
import scipy.ndimage
import astropy.units as u
import named_arrays as na
import xarray as xr

__all__ = [
    'ScalarStartT',
    'ScalarStopT',
    'ScalarTypeError',
    'as_named_array',
    'AbstractScalar',
    'AbstractScalarArray',
    'ScalarLike',
    'ScalarArray',
    'AbstractImplicitScalarArray',
    'ScalarUniformRandomSample',
    'ScalarNormalRandomSample',
    'ScalarPoissonRandomSample',
    'AbstractParameterizedScalarArray',
    'ScalarArrayRange',
    'AbstractScalarSpace',
    'ScalarLinearSpace',
    'ScalarStratifiedRandomSpace',
    'ScalarLogarithmicSpace',
    'ScalarGeometricSpace',
]

NDArrayT = TypeVar('NDArrayT', bound=npt.ArrayLike)
StartT = TypeVar('StartT', bound='ScalarLike')
StopT = TypeVar('StopT', bound='ScalarLike')
ScalarStartT = TypeVar('ScalarStartT', bound='ScalarLike')
ScalarStopT = TypeVar('ScalarStopT', bound='ScalarLike')
CenterT = TypeVar('CenterT', bound='ScalarLike')
WidthT = TypeVar('WidthT', bound='ScalarLike')
StartExponentT = TypeVar('StartExponentT', bound='ScalarLike')
StopExponentT = TypeVar('StopExponentT', bound='ScalarLike')
BaseT = TypeVar('BaseT', bound='ScalarLike')


class ScalarTypeError(TypeError):
    pass


def _normalize(a: float | u.Quantity | na.AbstractScalarArray) -> na.AbstractScalarArray:

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractScalarArray):
            result = a
        else:
            raise ScalarTypeError
    else:
        result = na.ScalarArray(a)

    return result


def as_named_array(value: bool | int | float | complex | str | u.Quantity | na.AbstractArray):
    if not hasattr(value, "__named_array_function__"):
        return ScalarArray(value)
    else:
        return value


@dataclasses.dataclass(eq=False, repr=False)
class AbstractScalar(
    na.AbstractArray,
):

    @property
    @abc.abstractmethod
    def dtype(self: Self) -> np.dtype:
        """
        Data type of the array. Equivalent to :attr:`numpy.ndarray.dtype`
        """

    @property
    def unit(self: Self) -> None | u.UnitBase:
        """
        Unit associated with the array.

        If :attr:`ndarray` is an instance of :class:`astropy.units.Quantity`, return :attr:`astropy.units.Quantity.unit`,
        otherwise return :class:`None`.
        """
        return na.unit(self)

    @property
    def unit_normalized(self: Self) -> u.UnitBase:
        """
        Similar to :attr:`unit` but returns :attr:`astropy.units.dimensionless_unscaled` if :attr:`ndarray` is not an
        instance of :class:`astropy.units.Quantity`.
        """
        return na.unit_normalized(self)

    @property
    def length(self) -> AbstractScalar:
        if np.issubdtype(self.dtype, np.number):
            return np.abs(self)
        else:
            raise ValueError('Can only compute length of numeric arrays')

    def volume_cell(self, axis: None | str | tuple[str]) -> na.AbstractScalar:
        if axis is None:
            if self.ndim != 1:
                raise ValueError(
                    f"If {axis=}, then {self.ndim=} must be one dimensional"
                )
            axis = self.axes[0]
        elif not isinstance(axis, str):
            axis, = axis

        return np.diff(self, axis=axis)

    def __array_matmul__(
            self: Self,
            x1: na.ArrayLike,
            x2: na.ArrayLike,
            out: None | AbstractScalar = None,
            **kwargs,
    ) -> na.AbstractExplicitArray:
        result = super().__array_matmul__(x1=x1, x2=x2, out=out, **kwargs)
        if result is not NotImplemented:
            return result

        if out is None:
            if np.all(x1 == 0) or np.all(x2 == 0):
                unit_1 = na.unit(x1, unit_dimensionless=1)
                unit_2 = na.unit(x2, unit_dimensionless=1)
                return 0 * unit_1 * unit_2

        return np.multiply(
            x1,
            x2,
            out=out,
            **kwargs,
        )

    def mean_trimmed(
        self,
        q: float = .25,
        axis: None | str | Sequence[str] = None,
    ) -> Self:
        """
        Compute the trimmed mean along the specified axes.

        Parameters
        ----------
        q
            The fraction of the largest and smallest elements to remove.
            Must be between 0 and 1/2.
            If the specified fraction does not result in an integer number of elements,
            the number of elements to trim is rounded down.
        axis
            The axis or axes along which to compute the trimmed mean.

        See Also
        --------
        :func:`mean_trimmed` functional version of this method.
        """
        return na.mean_trimmed(
            a=self,
            q=q,
            axis=axis,
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractScalarArray(
    AbstractScalar,
    Generic[NDArrayT],
):

    __named_array_priority__: ClassVar[int] = 1

    @property
    def type_abstract(self: Self) -> Type[AbstractScalarArray]:
        return AbstractScalarArray

    @property
    def type_explicit(self: Self) -> Type[ScalarArray]:
        return ScalarArray

    @property
    @abc.abstractmethod
    def ndarray(self: Self) -> bool | int | float | complex | str | np.ndarray | u.Quantity:
        """
        Underlying data that is wrapped by this class.

        This is usually an instance of :class:`numpy.ndarray` or :class:`astropy.units.Quantity`, but it can also be a
        built-in python type such as a :class:`int`, :class:`float`, or :class:`bool`
        """
    @property
    def to_xarray(self: Self) -> xr.DataArray:
        """
        Cast :class:`ScalarArray` to a :class:`xarray.DataArray`.  Useful for saving to netcdf/zarr or using Dask.
        """
        return xr.DataArray(self.ndarray, dims=self.axes)

    @property
    def value(self: Self) -> ScalarArray:
        return self.type_explicit(
            ndarray=na.value(self.ndarray),
            axes=self.axes,
        )

    def astype(
            self: Self,
            dtype: str | np.dtype | Type,
            order: str = 'K',
            casting='unsafe',
            subok: bool = True,
            copy: bool = True,
    ) -> ScalarArray:
        return ScalarArray(
            ndarray=np.asanyarray(self.ndarray).astype(
                dtype=dtype,
                order=order,
                casting=casting,
                subok=subok,
                copy=copy,
            ),
            axes=self.axes,
        )

    def to(
        self: Self,
        unit: u.Unit,
        equivalencies: None | list[tuple[u.Unit, u.Unit]] = [],
        copy: bool = True,
    ) -> ScalarArray:
        ndarray = self.ndarray
        if not isinstance(ndarray, u.Quantity):
            ndarray = ndarray << u.dimensionless_unscaled
        return ScalarArray(
            ndarray=ndarray.to(
                unit=unit,
                equivalencies=equivalencies,
                copy=copy,
            ),
            axes=self.axes,
        )

    def ndarray_aligned(self: Self, axes: Sequence[str]) -> np.ndarray:
        """
        Align :attr:`ndarray` to a particular sequence of axes.

        Parameters
        ----------
        axes
            New sequence of axes to align :attr:`ndarray` to.

        Returns
        -------
        An instance of :class:`numpy.ndarray` with the axes aligned.
        """
        axes = tuple(axes)
        axes_self = self.axes

        ndarray = self.ndarray
        if not hasattr(ndarray, "__array_function__"):
            ndarray = np.asanyarray(ndarray)

        if axes == axes_self:
            return ndarray

        ndim_missing = len(axes) - ndarray.ndim
        value = ndarray[(...,) + ndim_missing * (np.newaxis,)]
        source = []
        destination = []
        for axis_index, axis_name in enumerate(axes_self):
            source.append(axis_index)
            if axis_name not in axes:
                raise ValueError(
                    f"`axes` is missing axes present in the input array. "
                    f"`axes` is {axes} but `self.axes` is {self.axes}"
                )
            destination.append(axes.index(axis_name))
        value = np.moveaxis(value, source=source, destination=destination)
        return value

    def add_axes(self: Self, axes: str | Sequence[str]) -> ScalarArray:
        if isinstance(axes, str):
            axes = [axes]
        shape_new = {axis: 1 for axis in axes}
        shape = shape_new | self.shape
        return ScalarArray(
            ndarray=self.ndarray_aligned(shape),
            axes=tuple(shape.keys()),
        )

    def change_axis_index(self: Self, axis: str, index: int) -> ScalarArray:
        shape = self.shape
        size_axis = shape.pop(axis)
        keys = list(shape.keys())
        values = list(shape.values())
        index = index % len(self.shape) + 1
        keys.insert(index, axis)
        values.insert(index, size_axis)
        shape_new = {k: v for k, v in zip(keys, values)}
        return ScalarArray(
            ndarray=self.ndarray_aligned(shape_new),
            axes=tuple(keys),
        )

    def combine_axes(
            self: Self,
            axes: None | Sequence[str] = None,
            axis_new: None | str = None,
    ) -> ScalarArray:

        if axes is None:
            axes = self.axes

        if axis_new is None:
            axis_new = na.flatten_axes(axes)

        axes_new = list(self.axes)
        shape_new = self.shape
        for axis in axes:
            axes_new.append(axes_new.pop(axes_new.index(axis)))
            shape_new[axis] = shape_new.pop(axis)

        source = []
        destination = []
        for axis in axes:
            source.append(self.axes.index(axis))
            destination.append(axes_new.index(axis))

        for axis in axes:
            axes_new.remove(axis)
            shape_new.pop(axis)
        axes_new.append(axis_new)
        shape_new[axis_new] = -1

        return ScalarArray(
            ndarray=np.moveaxis(self.ndarray, source=source, destination=destination).reshape(tuple(shape_new.values())),
            axes=tuple(axes_new),
        )

    def to_string_array(
            self,
            format_value: str = "%.2f",
            format_unit: str = "latex_inline",
            pad_unit: str = r"$\,$"
    ):
        a = format_value

        unit = self.unit
        if unit is not None:
            a = f"{a}{pad_unit}{unit:{format_unit}}"

        return np.char.mod(
            a=a,
            values=self.value,
        )

    def _getitem(
            self: Self,
            item: dict[str, int | slice | AbstractScalarArray] | AbstractScalarArray,
    ):

        if isinstance(item, AbstractScalarArray):

            if not set(item.shape).issubset(self.axes):
                raise ValueError(
                    f"the axes in item, {item.axes}, must be a subset of the axes in the array, {self.axes}")

            value = np.moveaxis(
                a=self.ndarray,
                source=[self.axes.index(axis) for axis in item.axes],
                destination=np.arange(len(item.axes)),
            )

            if item.shape:
                axis_new = item.axes_flattened
            else:
                axis_new = "boolean"

            return ScalarArray(
                ndarray=np.moveaxis(value[item.ndarray], 0, ~0),
                axes=tuple(axis for axis in self.axes if axis not in item.axes) + (axis_new, )
            )

        elif isinstance(item, dict):

            axes = self.axes

            item = {ax: item[ax] for ax in item if ax in axes}

            if not item:
                if not axes:
                    ndarray = self.ndarray
                    if isinstance(ndarray, np.ndarray):
                        return self.replace(ndarray=self.ndarray[()])
                    else:
                        return self
                else:
                    return self

            item_advanced = dict()      # type: typ.Dict[str, AbstractScalarArray]
            for axis in item:
                item_axis = item[axis]
                if isinstance(item_axis, na.AbstractArray):
                    if isinstance(item_axis, AbstractScalarArray):
                        item_advanced[axis] = item_axis
                    else:
                        return NotImplemented
                elif isinstance(item_axis, int):
                    item_advanced[axis] = item_axis

            shape_advanced = na.shape_broadcasted(*item_advanced.values())

            ndarray_organized = np.moveaxis(
                self.ndarray,
                source=tuple(axes.index(ax) for ax in item_advanced),
                destination=tuple(range(len(item_advanced))),
            )

            axes_basic = tuple(ax for ax in axes if ax not in item_advanced)
            axes_self = tuple(item_advanced) + axes_basic
            axes_value = tuple(shape_advanced) + axes_basic

            index = [slice(None)] * self.ndim   # type: list[int | slice | AbstractScalar]
            for ax in item:
                item_axis = item[ax]
                if item_axis is None:
                    continue
                if isinstance(item_axis, na.AbstractScalarArray):
                    item_axis = item_axis.ndarray_aligned(shape_advanced)
                index[axes_self.index(ax)] = item_axis

            return ScalarArray(
                ndarray=ndarray_organized[tuple(index)],
                axes=axes_value,
            )

        else:
            return NotImplemented

    def _getitem_reversed(
            self: Self,
            array: AbstractScalarArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray,
    ):
        return NotImplemented

    def __bool__(self: Self) -> bool:
        return super().__bool__() and self.ndarray.__bool__()

    def __mul__(self: Self, other: na.ArrayLike | u.Unit) -> ScalarArray:
        if isinstance(other, u.UnitBase):
            return ScalarArray(
                ndarray=self.ndarray * other,
                axes=self.axes,
            )
        else:
            return super().__mul__(other)

    def __lshift__(self: Self, other: na.ArrayLike | u.UnitBase) -> ScalarArray:
        if isinstance(other, u.UnitBase):
            return ScalarArray(
                ndarray=self.ndarray << other,
                axes=self.axes
            )
        else:
            return super().__lshift__(other)

    def __truediv__(self: Self, other: na.ArrayLike | u.UnitBase) -> ScalarArray:
        if isinstance(other, u.UnitBase):
            return ScalarArray(
                ndarray=self.ndarray / other,
                axes=self.axes
            )
        else:
            return super().__truediv__(other)

    def __array_ufunc__(
            self,
            function: np.ufunc,
            method: str,
            *inputs,
            **kwargs,
    ) -> None | ScalarArray | tuple[ScalarArray, ...]:

        result = super().__array_ufunc__(
            function,
            method,
            *inputs,
            **kwargs,
        )
        if result is not NotImplemented:
            return result

        if function is np.matmul:
            return NotImplemented

        nout = function.nout

        kwargs_ndarray = kwargs.copy()

        if "out" in kwargs_ndarray:
            out = kwargs_ndarray["out"]
            out_normalized = list()
            for o in out:
                if o is not None:
                    if isinstance(o, ScalarArray):
                        if isinstance(o.ndarray, np.ndarray):
                            o = o.ndarray
                        else:
                            o = None
                    else:
                        return NotImplemented
                out_normalized.append(o)
            kwargs_ndarray["out"] = tuple(out_normalized)
        else:
            out = (None, ) * nout

        out_arrays = tuple(o for o in out if o is not None)
        if out_arrays:
            shape = out_arrays[0].shape
            if any(o.shape != shape for o in out_arrays[1:]):
                raise ValueError(
                    f"all the `out` arrays should have the same shape, got {tuple(o.shape for o in out_arrays)}"
                )
        else:
            shape = na.shape_broadcasted(*inputs)

        if 'where' in kwargs_ndarray:
            if isinstance(kwargs_ndarray['where'], na.AbstractArray):
                if isinstance(kwargs_ndarray['where'], na.AbstractScalarArray):
                    kwargs_ndarray['where'] = kwargs_ndarray['where'].ndarray_aligned(shape)
                else:
                    return NotImplemented

        inputs_ndarray = []
        for inp in inputs:
            if isinstance(inp, na.AbstractArray):
                if isinstance(inp, AbstractScalarArray):
                    inp = inp.ndarray_aligned(shape)
                else:
                    return NotImplemented
            elif inp is None:
                return None
            inputs_ndarray.append(inp)

        result_ndarray = getattr(function, method)(*inputs_ndarray, **kwargs_ndarray)
        if nout == 1:
            result_ndarray = (result_ndarray, )
        result = list(ScalarArray(result_ndarray[i], axes=tuple(shape.keys())) for i in range(nout))

        for i in range(nout):
            if out[i] is not None:
                out[i].ndarray = result[i].ndarray
                result[i] = out[i]

        if nout == 1:
            result = result[0]
        else:
            result = tuple(result)
        return result


    def __array_function__(
            self: Self,
            func: Callable,
            types: Collection,
            args: tuple,
            kwargs: dict[str, Any],
    ):
        result = super().__array_function__(func=func, types=types, args=args, kwargs=kwargs)
        if result is not NotImplemented:
            return result

        from . import scalar_array_functions

        if func in scalar_array_functions.SINGLE_ARG_FUNCTIONS:
            return scalar_array_functions.array_function_single_arg(func, *args, **kwargs)

        if func in scalar_array_functions.ARRAY_CREATION_LIKE_FUNCTIONS:
            return scalar_array_functions.array_function_array_creation_like(func, *args, **kwargs)

        if func in scalar_array_functions.SEQUENCE_FUNCTIONS:
            return scalar_array_functions.array_function_sequence(func, *args, **kwargs)

        if func in scalar_array_functions.DEFAULT_FUNCTIONS:
            return scalar_array_functions.array_function_default(func, *args, **kwargs)

        if func in scalar_array_functions.PERCENTILE_LIKE_FUNCTIONS:
            return scalar_array_functions.array_function_percentile_like(func, *args, **kwargs)

        if func in scalar_array_functions.ARG_REDUCE_FUNCTIONS:
            return scalar_array_functions.array_function_arg_reduce(func, *args, **kwargs)

        if func in scalar_array_functions.FFT_LIKE_FUNCTIONS:
            return scalar_array_functions.array_function_fft_like(func, *args, **kwargs)

        if func in scalar_array_functions.FFTN_LIKE_FUNCTIONS:
            return scalar_array_functions.array_function_fftn_like(func, *args, **kwargs)

        if func in scalar_array_functions.EMATH_FUNCTIONS:
            return scalar_array_functions.array_function_emath(func, *args, **kwargs)

        if func in scalar_array_functions.HANDLED_FUNCTIONS:
            return scalar_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def __named_array_function__(self, func, *args, **kwargs):
        result = super().__named_array_function__(func, *args, **kwargs)
        if result is not NotImplemented:
            return result

        from . import scalar_named_array_functions

        if func in scalar_named_array_functions.ASARRAY_LIKE_FUNCTIONS:
            return scalar_named_array_functions.asarray_like(func=func, *args, **kwargs)

        if func in scalar_named_array_functions.RANDOM_FUNCTIONS:
            return scalar_named_array_functions.random(func=func, *args, **kwargs)

        if func in scalar_named_array_functions.PLT_PLOT_LIKE_FUNCTIONS:
            return scalar_named_array_functions.plt_plot_like(func, *args, **kwargs)

        if func in scalar_named_array_functions.PLT_AXES_SETTERS:
            return scalar_named_array_functions.plt_axes_setter(func, *args, **kwargs)

        if func in scalar_named_array_functions.PLT_AXES_GETTERS:
            return scalar_named_array_functions.plt_axes_getter(func, *args, **kwargs)

        if func in scalar_named_array_functions.PLT_GET_LIM:
            return scalar_named_array_functions.plt_get_lim(func, *args, **kwargs)

        if func in scalar_named_array_functions.PLT_AXES_ATTRIBUTES:
            return scalar_named_array_functions.plt_axes_attribute(func, *args, **kwargs)

        if func in scalar_named_array_functions.NDFILTER_FUNCTIONS:
            return scalar_named_array_functions.ndfilter(func, *args, **kwargs)

        if func in scalar_named_array_functions.HANDLED_FUNCTIONS:
            return scalar_named_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def matrix_multiply(
            self: Self,
            other: AbstractScalar,
            axis_rows: str,
            axis_columns: str,
    ) -> ScalarArray:

        shape = self.shape_broadcasted(other)
        shape_rows = shape.pop(axis_rows)
        shape_columns = shape.pop(axis_columns)
        shape = {**shape, axis_rows: shape_rows, axis_columns: shape_columns}

        data_self = self.ndarray_aligned(shape)
        data_other = other.ndarray_aligned(shape)

        return ScalarArray(
            ndarray=np.matmul(data_self, data_other),
            axes=list(shape.keys()),
        )

    def matrix_determinant(
            self: Self,
            axis_rows: str,
            axis_columns: str
    ) -> ScalarArray:
        shape = self.shape
        if shape[axis_rows] != shape[axis_columns]:
            raise ValueError('Matrix must be square')

        if shape[axis_rows] == 2:
            a = self[{axis_rows: 0, axis_columns: 0}]
            b = self[{axis_rows: 0, axis_columns: 1}]
            c = self[{axis_rows: 1, axis_columns: 0}]
            d = self[{axis_rows: 1, axis_columns: 1}]
            return a * d - b * c

        elif shape[axis_rows] == 3:
            a = self[{axis_rows: 0, axis_columns: 0}]
            b = self[{axis_rows: 0, axis_columns: 1}]
            c = self[{axis_rows: 0, axis_columns: 2}]
            d = self[{axis_rows: 1, axis_columns: 0}]
            e = self[{axis_rows: 1, axis_columns: 1}]
            f = self[{axis_rows: 1, axis_columns: 2}]
            g = self[{axis_rows: 2, axis_columns: 0}]
            h = self[{axis_rows: 2, axis_columns: 1}]
            i = self[{axis_rows: 2, axis_columns: 2}]
            return (a * e * i) + (b * f * g) + (c * d * h) - (c * e * g) - (b * d * i) - (a * f * h)

        else:
            value = np.moveaxis(
                a=self.ndarray,
                source=[self.axis_names.index(axis_rows), self.axis_names.index(axis_columns)],
                destination=[~1, ~0],
            )

            axes_new = list(self.axes)
            axes_new.remove(axis_rows)
            axes_new.remove(axis_columns)

            return ScalarArray(
                ndarray=np.linalg.det(value),
                axes=tuple(axes_new),
            )

    def matrix_inverse(
            self: Self,
            axis_rows: str,
            axis_columns: str,
    ) -> ScalarArray:

        axis_rows_inverse = axis_columns
        axis_columns_inverse = axis_rows

        index_axis_rows = self.axes.index(axis_rows)
        index_axis_columns = self.axes.index(axis_columns)
        value = np.moveaxis(
            a=self.ndarray,
            source=[index_axis_rows, index_axis_columns],
            destination=[~1, ~0],
        )

        axes_new = list(self.axes)
        axes_new.remove(axis_rows)
        axes_new.remove(axis_columns)
        axes_new.append(axis_rows_inverse)
        axes_new.append(axis_columns_inverse)

        return ScalarArray(
            ndarray=np.linalg.inv(value),
            axes=tuple(axes_new),
        )

    def filter_median(
            self: Self,
            shape_kernel: dict[str, int],
            mode: str = 'reflect',
    ):

        shape = self.shape
        shape_kernel_final = {axis: shape_kernel[axis] if axis in shape_kernel else 1 for axis in shape}

        inp = self.ndarray
        if isinstance(inp, u.Quantity):
            inp = inp.value
        result = ScalarArray(
            ndarray=scipy.ndimage.median_filter(
                input=inp,
                size=tuple(shape_kernel_final.values()),
                mode=mode,
            ),
            axes=self.axes,
        )
        if isinstance(self.ndarray, u.Quantity):
            result = result << self.unit

        return result


ScalarLike = Union[na.QuantityLike, AbstractScalar]


@dataclasses.dataclass(eq=False, repr=False)
class ScalarArray(
    AbstractScalarArray,
    na.AbstractExplicitArray,
    Generic[NDArrayT],
):
    """
    An array representing a scalar quantity (like pressure or temperature) with names for each of its `N` axes.

    A :class:`ScalarArray` is defined by a :class:`numpy.ndarray` and a :class:`tuple` of axis names.

    .. jupyter-execute::

        import numpy as np
        import named_arrays as na

        x = np.array([1, 2, 3])
        x = na.ScalarArray(x, axes=('position_x', ))
        print(x)

    They can also be defined using a :class:`astropy.units.Quantity`, or ascribed units after creation.

    .. jupyter-execute::

        import astropy.units as u

        x = x * u.cm

        print(x)

    What happens when we do math with instances of :class:`ScalarArray`?

    .. jupyter-execute::

        y = np.array([4, 5, 6]) * u.mm

        y = na.ScalarArray(y, axes=('position_y', ))
        print(y**2)

        radius = np.sqrt(x**2 + y**2)
        print(radius)
        print(radius.shape)

    Note how when performing mathematical operations on two instances of :class:`ScalarArray` with different axes, the
    arrays are automatically broadcast over every axis. There is no need to insert extra dimensions for alignment like
    you would normally do with instances of :class:`numpy.ndarray`.

    We can also use reduction operators (mean, sum, etc) on instances of :class:`ScalarArray` by using the axis name
    instead of the axis index.

    .. jupyter-execute::

        print(radius.mean())
        print(radius.mean(axis='position_x'))
    """

    ndarray: None | NDArrayT = 0
    axes: None | str | tuple[str, ...] = None

    def __post_init__(self: Self):
        if self.axes is None:
            self.axes = tuple()
        if isinstance(self.axes, str):
            self.axes = (self.axes, )
        if getattr(self.ndarray, 'ndim', 0) != len(self.axes):  # pragma: nocover
            raise ValueError(
                f'The number of axis names, {self.axes}, '
                f'must match the number of dimensions, {np.ndim(self.ndarray)}.'
            )
        if len(self.axes) != len(set(self.axes)):
            raise ValueError(f'Each axis name must be unique, got {self.axes}.')

    @classmethod
    def from_scalar_array(
            cls: type[Self],
            a: float | u.Quantity | na.AbstractScalarArray,
            like: None | Self = None,
    ) -> Self:

        self = super().from_scalar_array(a=a, like=like)

        if isinstance(a, na.AbstractArray):
            if isinstance(a, na.AbstractScalarArray):
                self.ndarray = a.ndarray
                self.axes = a.axes
            else:
                raise TypeError(
                    f"If `a` is an instance of `{na.AbstractArray.__name__}`, it must be an instance of "
                    f"`{na.AbstractScalarArray.__name__}`, got `{type(a).__name__}`."
                )
        else:
            self.ndarray = a
            self.axes = tuple()

        return self

    @classmethod
    def from_xarray(cls, dataarray: xr.DataArray) -> Self:
        """
        Convert :class:`xarray.DataArray` into a :class:`ScalarArray`
        """
        return cls(dataarray.data, axes=dataarray.dims)

    @classmethod
    def empty(cls: Type[Self], shape: dict[str, int], dtype: Type | np.dtype = float) -> Self:
        """
        Create a new empty array

        Parameters
        ----------
        shape
            shape of the new array
        dtype
            data type of the new array

        Returns
        -------
            A new empty array with the specified shape and data type
        """
        return cls(
            ndarray=np.empty(shape=tuple(shape.values()), dtype=dtype),
            axes=tuple(shape.keys()),
        )

    @classmethod
    def zeros(cls: Type[Self], shape: dict[str, int], dtype: Type | np.dtype = float) -> Self:
        """
        Create a new array of zeros

        Parameters
        ----------
        shape
            shape of the new array
        dtype
            data type of the new array

        Returns
        -------
            A new array of zeros with the specified shape and data type
        """
        return cls(
            ndarray=np.zeros(shape=tuple(shape.values()), dtype=dtype),
            axes=tuple(shape.keys()),
        )

    @classmethod
    def ones(cls: Type[Self], shape: dict[str, int], dtype: Type | np.dtype = float) -> Self:
        """
        Create a new array of ones

        Parameters
        ----------
        shape
            shape of the new array
        dtype
            data type of the new array

        Returns
        -------
            A new array of ones with the specified shape and data type
        """
        return cls(
            ndarray=np.ones(shape=tuple(shape.values()), dtype=dtype),
            axes=tuple(shape.keys()),
        )

    @property
    def shape(self: Self) -> dict[str, int]:
        try:
            return dict(zip(self.axes, self.ndarray.shape, strict=True))
        except AttributeError:
            return dict()

    @property
    def ndim(self: Self) -> int:
        return np.ndim(self.ndarray)

    @property
    def size(self: Self) -> int:
        return np.size(self.ndarray)

    @property
    def dtype(self: Self) -> np.dtype:
        return na.get_dtype(self.ndarray)

    @property
    def explicit(self: Self) -> Self:
        return self

    def __setitem__(
            self: Self,
            item: dict[str, int | slice | AbstractScalarArray] | AbstractScalarArray,
            value: int | float | u.Quantity | AbstractScalarArray,
    ) -> None:

        shape_self = self.shape

        if isinstance(value, na.AbstractArray):
            if isinstance(value, na.AbstractScalarArray):
                value = value.explicit
            else:
                raise TypeError(
                    f"if `value` is an instance of `AbstractArray`, it must be an instance of `AbstractScalarArray`, "
                    f"got {type(value)}"
                )
        else:
            value = ScalarArray(value)

        if isinstance(item, AbstractScalarArray):

            item = item.explicit
            shape_item = item.shape

            if not set(item.shape).issubset(shape_self):
                raise ValueError(
                    f"if `item` is an instance of `{na.AbstractArray.__name__}`, "
                    f"`item.axes`, {item.axes}, should be a subset of `self.axes`, {self.axes}"
                )

            if shape_item:
                axis_new = item.axes_flattened
            else:
                axis_new = "boolean"

            axes_untouched = tuple(ax for ax in shape_self if ax not in shape_item)
            axes_value = (axis_new, ) + axes_untouched
            axes_self = tuple(shape_item) + axes_untouched

            self.ndarray_aligned(axes_self)[item.ndarray] = value.ndarray_aligned(axes_value)

        elif isinstance(item, dict):

            if not set(item).issubset(shape_self):
                raise ValueError(
                    f"{item.keys()=} must be a subset of {self.axes=}"
                )

            item_advanced = dict()  # type: typ.Dict[str, AbstractScalarArray]
            for axis in item:
                item_axis = item[axis]
                if isinstance(item_axis, na.AbstractArray):
                    if isinstance(item_axis, AbstractScalarArray):
                        item_advanced[axis] = item_axis
                    else:  #pragma: nocover
                        raise TypeError(
                            "if `item[axis]` is an instance of `AbstractArray`, "
                            "it must be an instance of `AbstractScalarArray`, "
                            f"got {type(item[axis])=}."
                        )
                elif isinstance(item_axis, int):
                    item_advanced[axis] = item_axis

            shape_advanced = na.shape_broadcasted(*item_advanced.values())

            axes_basic = tuple(ax for ax in shape_self if ax not in item_advanced)
            axes_self = tuple(item_advanced) + axes_basic
            axes_value = tuple(shape_advanced) + axes_basic

            index = [slice(None)] * len(axes_self)   # type: list[Union[int, slice, AbstractScalar]]
            for axis in item:
                item_axis = item[axis]
                if isinstance(item_axis, na.AbstractScalarArray):
                    item_axis = item_axis.ndarray_aligned(shape_advanced)
                index[axes_self.index(axis)] = item_axis

            if value.shape:
                value = value.ndarray_aligned(axes_value)
            else:
                value = value.ndarray

            self.ndarray_aligned(axes_self)[tuple(index)] = value


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitScalarArray(
    AbstractScalarArray,
    na.AbstractImplicitArray,
):

    @property
    def ndarray(self: Self) -> na.QuantityLike:
        return self.explicit.ndarray

    @property
    def dtype(self: Self) -> np.dtype:
        return self.explicit.dtype

    def _attr_normalized(self, name: str) -> ScalarArray:

        attr = getattr(self, name)

        if isinstance(attr, na.AbstractArray):
            if isinstance(attr, na.AbstractScalarArray):
                result = attr
            else:
                raise TypeError(
                    f"if `{name}` is an instance of `AbstractArray`, it must be an instance of `AbstractScalarArray`, "
                    f"got {type(attr)}"
                )
        else:
            result = ScalarArray(attr)

        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractScalarRandomSample(
    AbstractImplicitScalarArray,
    na.AbstractRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ScalarUniformRandomSample(
    AbstractScalarRandomSample,
    na.AbstractUniformRandomSample[ScalarStartT, ScalarStopT],
):
    def volume_cell(self, axis: None | str | tuple[str]) -> na.AbstractScalar:
        axis = na.axis_normalized(self, axis)
        if len(axis) != 1:
            raise ValueError(
                f"{axis=} must have exactly one element for scalars."
            )
        axis, = axis

        shape_random = self.shape_random
        if axis in shape_random:
            result = (self.stop - self.start) / shape_random[axis]
        else:
            result = super().volume_cell(axis)

        return result


@dataclasses.dataclass(eq=False, repr=False)
class ScalarNormalRandomSample(
    AbstractScalarRandomSample,
    na.AbstractNormalRandomSample[CenterT, WidthT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ScalarPoissonRandomSample(
    AbstractScalarRandomSample,
    na.AbstractPoissonRandomSample[CenterT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedScalarArray(
    AbstractImplicitScalarArray,
    na.AbstractParameterizedArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ScalarArrayRange(
    AbstractParameterizedScalarArray,
    na.AbstractArrayRange[ScalarStartT, ScalarStopT],
):
    """
    An :class:`AbstractScalarArray` over the range [:attr:`start`, :attr:`stop`) incremented by :attr:`step`.
    An analog to :func:`numpy.arange`.

    :class:`ScalarArrayRange` can be used to create a :class:`ScalarArray` of integers.

    .. jupyter-execute::

        import named_arrays as na
        x = na.ScalarArrayRange(1, 8, axis = "x")
        print(x.explicit)
        print(x.shape)

    Note above that ``x`` does not include :attr:`stop`, and won't in almost all cases.  :class:`ScalarArrayRange` can be used to
    create an increasing :class:`ScalarArray` of floats, even with non integer steps.

    .. jupyter-execute::

        x = na.ScalarArrayRange(-0.5, 3, "x", 0.25)
        print(x.explicit)

    For the above, and more complicated uses, it is recommended to use :class:`ScalarLinearSpace` instead.  See numpy
    documentation of :func:`numpy.arange` for more info.
    """
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractScalarSpace(
    AbstractParameterizedScalarArray,
    na.AbstractSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ScalarLinearSpace(
    AbstractScalarSpace,
    na.AbstractLinearSpace[StartT, StopT, str, int],
):
    """
    An evenly spaced :class:`ScalarArray` ranging from start to stop with num elements.

    Most often, instances of :class:`ScalarArray` won't be formed directly from a :class:`numpy.ndarray`, but through
    more useful routines like :class:`ScalarLinearSpace`, a named_arrays equivalent to :func:`numpy.linspace`.
    For example, one can quickly create an evenly spaced coordinate (or axis) array with units.

    .. jupyter-execute::

        import named_arrays as na
        import astropy.units as u

        photon_energy = na.ScalarLinearSpace(1, 10, axis="energy", num=10) * u.keV
        print(photon_energy.shape)
        print(photon_energy)

    Then easily convert that into a wavelength.

    .. jupyter-execute::

        wavelength = (1240 * u.eV * u.nm / photon_energy).to(u.nm)
        print(wavelength)
    """

    def volume_cell(self, axis: None | str | tuple[str]) -> na.AbstractScalar:
        axis = na.axis_normalized(self, axis)
        if len(axis) != 1:
            raise ValueError(
                f"{axis=} must have exactly one element for scalars."
            )
        axis, = axis

        if axis == self.axis:
            result = self.step

        else:
            result = super().volume_cell(axis)

        return result

    # def index(
    #         self: ScalarLinearSpaceT,
    #         value: ScalarLinearSpaceT,
    #         axis: None | str | Sequence[str] = None,
    # ) -> dict[str, ScalarArrayT]:
    #     result = super().index(
    #         value=value,
    #         axis=[a for a in axis if a != self.axis],
    #     )
    #     result[self.axis] = (value - self.start) / self.step
    #
    #     return result
    #
    # def index_nearest(self, value: AbstractScalarT) -> typ.Dict[str, AbstractScalarT]:
    #     return {self.axis: np.rint((value - self.start) / self.step).astype(int)}
    #
    # def index_below(self, value: AbstractScalarT) -> typ.Dict[str, AbstractScalarT]:
    #     return {self.axis: (value - self.start) // self.step}


@dataclasses.dataclass(eq=False, repr=False)
class ScalarStratifiedRandomSpace(
    ScalarLinearSpace[StartT, StopT],
    na.AbstractStratifiedRandomSpace[StartT, StopT, str, int],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ScalarLogarithmicSpace(
    AbstractScalarSpace,
    na.AbstractLogarithmicSpace[StartExponentT, StopExponentT, BaseT, str, int],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ScalarGeometricSpace(
    AbstractScalarSpace,
    na.AbstractGeometricSpace[StartT, StopT, str, int],
):
    pass
